#include <stdlib.h>
#include <stdio.h>
#include <windows.h>
#include <assert.h>
#include <math.h>

//char * test_bmp = "C:\\Documents and Settings\\greenwbr\\Desktop\\test2.bmp";

void WriteBmp(const unsigned char * const in,char * filename, int width, int height ){
	unsigned char header[54];
	memset(header,0,54);

	header[0]='B';
	header[1]='M';

	int stride = (width*3+3)&(~3);

	*(long*)(header+2)=height*stride+54;
	*(long*)(header+10)=54;
	*(long*)(header+14)=0x28;
	*(long*)(header+18)=width;
	*(long*)(header+22)=height;
	*(short*)(header+26)=1;
	*(short*)(header+28)=24;
	*(long*)(header+34)=height*stride;
	
	HANDLE file = CreateFileA(filename,GENERIC_WRITE,FILE_SHARE_READ|FILE_SHARE_WRITE,NULL,CREATE_ALWAYS,0,NULL);
	unsigned long bytes;
	WriteFile(file, header,54,&bytes,NULL);
	if ( stride == width*3){
		WriteFile(file, in, width*height*3, &bytes, NULL );
	} else {
		int zero=0;
		for ( int h=0;h<height;h++){
			WriteFile(file, in+width*3*h, width*3, &bytes, NULL );
			WriteFile(file, &zero, stride-width*3, &bytes, NULL );
		}
	}
	
	CloseHandle(file);
}

int ReadBMP(char * filename, int &width, int &height, unsigned char * &data){
	unsigned char header[54];
	HANDLE file = CreateFileA( filename, GENERIC_READ,FILE_SHARE_READ,NULL,OPEN_EXISTING,NULL,NULL);
	if ( file == (void*)0xFFFFFFFF )
		return 1;
	unsigned long fsize = GetFileSize(file,NULL);
	if ( fsize < 54 )
		return 2;
	unsigned long bytes;
	ReadFile(file,header,54,&bytes,NULL);
	width = *(int*)(header+0x12);
	height =*(int*)(header+0x16);
	if ( !(width & 3) ){
		data = (unsigned char *)malloc(width*height*4);
		ReadFile(file,data,width*height*3,&bytes,NULL);
		CloseHandle(file);
	} else {
		data = (unsigned char *)malloc(width*height*4);
		unsigned char * unaligned =(unsigned char *)malloc(fsize-54);
		ReadFile(file,unaligned,fsize-54,&bytes,NULL);
		CloseHandle(file);
		for ( int h=0;h<height;h++){
			memcpy(data+h*width*3,unaligned+h*((width*3+3)&(~3)),width*3);
		}
		free(unaligned);
	}
	

	return 0;
}

#define TOP_VALUE		0x80000000
#define BOTTOM_VALUE	0x00800000
#define SHIFT_BITS		23


#define output_byte( x ) *are_out++=x;

unsigned int are_low = 0;
unsigned int are_range = TOP_VALUE;
unsigned int are_buff = 0;
unsigned int are_help = 0;
unsigned int are_frequency[256];
int are_ptotal = 16;
unsigned char * are_out=NULL;

void init_range_encode(unsigned char * buffer, int ptotal){
	are_ptotal=ptotal;
	for ( int a=0;a<are_ptotal;a++){
		are_frequency[a]=1;
	}
	are_low = 0;
	are_range = TOP_VALUE;
	are_buff = 0;
	are_help = 0;
	are_ptotal = 16;
	are_out=buffer;
}


void adaptive_range_encode( int invalue){

	// encode symbol
	unsigned int r = are_range /are_ptotal;
	unsigned int tmp=0;
	for ( int a=0;a<invalue;a++)
		tmp+=are_frequency[a];
	tmp *= r;
	are_low += tmp;
	if ( invalue < 255 ) {
		are_range = r * (are_frequency[invalue]);
		if ( are_range == 0){
			printf("error\n");
		}
	}else {
		are_range -= tmp;	
		if ( are_range == 0){
			printf("error\n");
		}
	}
	are_frequency[invalue]++;
	are_ptotal++;

	while ( are_range <= BOTTOM_VALUE )  {
		if ( (are_low>>SHIFT_BITS) < 255  ) {
			output_byte ( are_buff );
			are_buff = are_low >> SHIFT_BITS;
			for ( ; are_help != 0 ; are_help-- ){
				output_byte ( 255 );
			}
		} else { 
			if ( are_low & TOP_VALUE ) {
				output_byte( are_buff + 1 );
				are_buff = are_low >> SHIFT_BITS;
				for ( ; are_help ; are_help-- ){
					output_byte ( 0 );
				}
			} else {
				are_help++;
			}
		}
		are_range <<= 8;
		are_low <<= 8;
		are_low &= (TOP_VALUE-1);
	}
}

void flush_are(){
	if ( are_low >> 23 > 0xFF ){
		output_byte( are_buff+1 );
		while ( are_help-- )
			*are_out++=0;
	} else {
		output_byte ( are_buff );
		while ( are_help-- )
			*are_out++=255;
	}
	output_byte( are_low>>23 );
	output_byte( are_low>>15 );
	output_byte( are_low>>7 );
	output_byte( are_low<<1 );
}

#define inbyte() ((unsigned int)(*ard_in++))

unsigned int ard_range = 0;
unsigned int ard_low = 0;
unsigned int ard_buff = 0;
unsigned int ard_frequency[] = {1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1};
int ard_ptotal = 16;
unsigned char * ard_in=NULL;

void init_range_decode(unsigned char * buffer, int ptotal){
	ard_ptotal=ptotal;
	for ( int a=0;a<ard_ptotal;a++){
		ard_frequency[a]=1;
	}
	ard_in=buffer;
	ard_low = inbyte();
	ard_range=0x80;
	ard_buff = 0;
	
}

int adaptive_range_decode(){
  
	int ret_val=0;

	while( ard_range <= 0x00800000 ){
		ard_low <<= 8;
		ard_low += inbyte();
		ard_range <<= 8;
	}
	
	unsigned int help = ard_range / ard_ptotal;
	unsigned int tmp = ard_low/help;

	int x=0;
	int base=0;
	do {
		base+=ard_frequency[x];
		x++;
	} while ( base <= tmp);
	x--;
	//for (; p_ranges[x+1] <= tmp;x++);

	ret_val = x;

	ard_low -= help * (base-ard_frequency[x]);
	ard_range = help * (ard_frequency[x]);

	ard_frequency[x]++;
	ard_ptotal++;
		
	return ret_val;

}

void downsize(const unsigned char *const rgb,unsigned char *tmp, const int ow, const int oh, const int nw, const int nh){

	double ws = (ow-1)/(double)nw;
	double hs = (oh-1)/(double)nh;
	for ( int h=0;h<nh;h++){
		for ( int w=0;w<nw;w++){
			for ( int p = 0;p<3;p++){
				double sum=0;
				double hl = h*hs;
				double hh = hl+hs;
				double wl = w*ws;
				double wh = wl+ws;
			
				for ( int b=floor(hl);b<ceil(hh);b++){
					for ( int a=floor(wl);a<ceil(wh);a++){
						if ( b >= oh || a>=ow){
							printf("overrrun\n");
						}
						double v = rgb[b*ow*3+a*3+p];
						if ( b < hl ){
							v *= ceil(hl)-hl;
						}
						if ( b == floor(hh) ){
							v *= hh-b;
						}
						if ( a < wl ){
							v *= ceil(wl)-wl;
						}
						if ( a == floor(wh) ){
							v *= wh-a;
						}
						sum+=v;
					}
				}
				sum/= ws*hs;
				sum += 0.5;
				tmp[h*nw*3+w*3+p]=min(max(sum,0),255);
			}
		}
	}
}

int floor2(double d){ return (int)d;}
int ceil2(double d){ return (int)(d+1);}
int min2(int a,int b){
	return (int)min(a,b-1);
}

void enlarge(unsigned char * src, unsigned char * dst, int iw,int ih, int width, int height){
	for ( int c=0;c<3;c++){
		for ( int h=0;h<height;h++){
			for ( int w=0;w<width;w++){
				double px = w/(double)width*iw;
				double py = h/(double)height*ih;

				double sum = 0;
				sum += src[(int)(floor2(py)*iw*3+floor2(px)*3+c)] * (ceil2(px)-px)*(ceil2(py)-py);
				sum += src[(int)(floor2(py)*iw*3+min2(ceil2(px),iw)*3+c)] * (px-floor2(px))*(ceil2(py)-py);
				sum += src[(int)(min2(ceil2(py),ih)*iw*3+min2(ceil2(px),iw)*3+c)] * (px-floor2(px))*(py-floor2(py));
				sum += src[(int)(min2(ceil2(py),ih)*iw*3+floor2(px)*3+c)] * (ceil2(px)-px)*(py-floor2(py));
				dst[h*width*3+w*3+c]=max(min(sum+0.5,255),0);
			}
		}
	}
}


void double_res(unsigned char * src, int ow, int oh){
	
	double * tmp = new double[ow*2*oh*2];

	for ( int c=0;c<3;c++){
		for ( int h=0;h<oh*2;h++){
			for ( int w=0;w<ow*2;w++){
				tmp[h*ow*2+w] = src[h/2*ow*3+w/2*3+c];
			}
		}

		for ( int h=0;h<oh*2;h++){
			for ( int w=1;w<ow*2-1;w+=2){
				double d0 = tmp[h*ow*2+w];
				double d1 = tmp[h*ow*2+w+1];
				tmp[h*ow*2+w] = d0*0.75+d1*0.25;
				tmp[h*ow*2+w+1]= d0*0.25+d1*0.75;
			}
		}
		for ( int h=1;h<oh*2-2;h+=2){
			for ( int w=0;w<ow*2;w++){
				double d0 = tmp[h*ow*2+w];
				double d1 = tmp[(h+1)*ow*2+w];
				tmp[h*ow*2+w] = d0*0.75+d1*0.25;
				tmp[(h+1)*ow*2+w]= d0*0.25+d1*0.75;
			}
		}
		for ( int a=0;a<oh*2*ow*2;a++){
			src[a*3+c]=(int)(tmp[a]+0.5);
		}
	}
	delete [] tmp;

}

void RGB2YUV(unsigned char * src, unsigned char * dst, int width, int height){
	for ( int h=0;h<height;h++){
		for ( int w=0;w<width;w++){
			int p = h*width*3+w*3;
			int b = src[p+0];
			int g = src[p+1];
			int r = src[p+2];
			double y = 0.299*r + 0.587*g + 0.114*b;
			double u = (b-y)*0.565+128;
			double v = (r-y)*0.713+128;
			
			dst[p+0]=(unsigned char)min(max(y+0.5,0),255);
			dst[p+1]=(unsigned char)min(max(u+0.5,0),255);
			dst[p+2]=(unsigned char)min(max(v+0.5,0),255);
		}
	}
}

void YUV2RGB(unsigned char * src, unsigned char * dst, int width, int height){
	for ( int h=0;h<height;h++){
		for ( int w=0;w<width;w++){
			int p = h*width*3+w*3;
			int y = src[p+0];
			int u = src[p+1]-128;
			int v = src[p+2]-128;
			
			double r = y + 1.403*v;
			double g = y - 0.344*u - 0.714*v;
			double b = y + 1.770*u;
			
			dst[p+0]=(unsigned char)min(max(b+0.5,0),255);
			dst[p+1]=(unsigned char)min(max(g+0.5,0),255);
			dst[p+2]=(unsigned char)min(max(r+0.5,0),255);
		}
	}
}

void add(unsigned char * data, int value){
	for ( int a=0;a<=114&&value>0;a++){
		value += data[a];
		data[a]=value;
		value>>=8;
	}
}

void multiply(unsigned char * data, int value){
	for ( int a=114;a>=0;a--){
		int carry = data[a]*value;
		data[a]=carry;
		carry >>=8;
		
		int i=a+1;
		while ( carry ){
			carry += data[i];
			data[i]=carry;
			carry >>= 8;
			i++;
		}
	}
}

int divide(unsigned char * data, int value){
	int carry=0;
	for ( int a=114;a>=0;a--){
		int v = carry*256 + data[a];
		carry = v%value;
		data[a]=v/value;
	}
	return carry;
}

char temp_tweet[256];
void output_tweet(unsigned char * data){

	memset(temp_tweet,32,256);
	int range = 127-32;
	for ( int a=0;a<140;a++){
		int c = divide(data,range);
		c+=32;
		printf("%c",c);
		temp_tweet[a]=c;
	}
	printf("\n");
}

void input_tweet(char * text, unsigned char * data){
	
	memset(data,0,256);
	int range = 127-32;
	for ( int a=139;a>=0;a--){
		multiply(data,range);
		assert(text[a]>=32 && text[a] <= 126);
		add(data,text[a]-32);
	}
}

inline int median(int x,int y,int z ) {
	int delta = x-y;
	delta &= delta>>31;
	y+=delta; // min
	x-=delta; // max
	delta = y-z;
	delta &= delta>>31;
	y -=delta;	// max
	delta = y-x;
	delta &= delta>>31;
	return  x+delta;	// min
}

void predict(unsigned char * src, int width, int height){
	for ( int a=0;a<width*height*3;a++){
		src[a]>>=4;
	}
	int stride = width*3;
	for ( int c=0;c<3;c++){
		for ( int h=height-1;h>0;h--){
			for ( int w=width-1;w>0;w--){
				int p = h*stride+w*3+c;
				int x = src[p-3];
				int y = src[p-stride];
				int z = src[p-stride-3];
				src[p]-=median(x,y,x+y-z);
				src[p]&=0xf;
			}
			src[h*stride+c]-=src[(h-1)*stride+c];
		}
		for ( int w=width-1;w>0;w--){
			src[w*3+c]-=src[w*3-3+c];
		}
	}
	for ( int a=0;a<width*height*3;a++){
		src[a]&=0xf;
	}
}

void restore(unsigned char * src, int width, int height){

	for ( int c=0;c<3;c++){

		for ( int w=1;w<width;w++){
			src[w*3+c]+=src[w*3-3+c];
			src[w*3+c]&=0xf;
		}

		int stride = width*3;
		for ( int h=1;h<height;h++){
			src[h*stride+c]+=src[(h-1)*stride+c];
			src[h*stride+c]&=0xf;
			for ( int w=1;w<width;w++){
				int p = h*stride+w*3+c;
				int x = src[p-3];
				int y = src[p-stride];
				int z = src[p-stride-3];
				src[p]+=median(x,y,x+y-z);
				src[p]&=0xf;
			}
		}
		
	}
	for ( int a=0;a<width*height*3;a++){
		src[a]<<=4;
	}
}

int encode(const unsigned char * const src,unsigned char * buffer,unsigned char * tweet,int owidth,int oheight,int nwidth,int nheight, bool color){
	init_range_encode(tweet,16);
	downsize(src,buffer,owidth,oheight,nwidth,nheight);
	predict(buffer,nwidth,nheight);
	for ( int a=0;a<nwidth*nheight;a++){
		adaptive_range_encode(buffer[a*3]);
	}

	if ( color ){
		downsize(src,buffer,owidth,oheight,nwidth/3,nheight/3);
		predict(buffer,nwidth/3,nheight/3);
		for ( int a=0;a<(nwidth/3)*(nheight/3);a++){
			adaptive_range_encode(buffer[a*3+1]);
			adaptive_range_encode(buffer[a*3+2]);
		}
	}

	flush_are();
	return are_out-tweet;
}

void compress(char * filename){
	
	unsigned char tweet[1024*32];
	
	unsigned char * data;
	int width, height;
	if ( ReadBMP(filename,width,height,data) != 0 ){
		printf("Cannot open image file \"%s\"\n",filename);
		exit(0);
	}
	
	unsigned char * buffer = new unsigned char[width*height*3];
	unsigned char * buffer2 = new unsigned char[width*height*3];
	unsigned char * buffer3 = new unsigned char[width*height*3];

	unsigned char * uv = new unsigned char[width*height*3];

	RGB2YUV(data,buffer,width,height);
	
	downsize(buffer,buffer2,width,height,16,16);
	int c=0;
	for ( int h=0;h<16;h++){
		for ( int w=0;w<16;w++){
			c+=abs(buffer2[h*16*3+w*3+1]-128)/16;
			c+=abs(buffer2[h*16*3+w*3+2]-128)/16;
		}
	}
	bool color = c>2;

	int bw=4;
	int bh=4;
	int bs=0;

	double ar = height;
	ar/=width;

	for ( int w=8;w<32;w++){
		
		int h = w*ar;
		
		memset(tweet,0,sizeof(tweet));
		int size = encode(buffer,buffer2,tweet+1,width,height,w,h,color);
		
		if ( size <=112){
			bs = size;
			bw = w;
			bh = h;
		}

		memset(tweet,0,sizeof(tweet));
		size = encode(buffer,buffer2,tweet+1,width,height,w,h+1,color);
		
		if ( size <=112){
			bs = size;
			bw = w;
			bh = h+1;
		}
	}

	memset(tweet,0,sizeof(tweet));
	
	encode(buffer,buffer2,tweet+2,width,height,bw,bh,color);

	unsigned int b=0;
	for ( int a=3;a<114;a++){
		b<<=8;
		b|=tweet[a];
		tweet[a-1] = (b>>1)&0xff;
	}

	if ( color ){
		tweet[0]=bw;
	} else {
		tweet[0]=-bw;
	}
	
	tweet[1]=bh;

	memset(tweet+114,0,128);

	output_tweet(tweet);
}

void decompress(char * tweet_file, char * image_file){

	unsigned char tweet[1024];
	char text[256];
	if ( tweet_file != NULL ){
		HANDLE file = CreateFileA(tweet_file,GENERIC_READ,0,0,OPEN_EXISTING,0,0);
		if ( file == (HANDLE)-1 || file == (HANDLE)0){
			printf("Cannot open text file \"%s\"\n",tweet_file);
			exit(0);
		}
		unsigned long bytes;
		ReadFile(file,text,140,&bytes,0);
		CloseHandle(file);
	} else {
		memcpy(text,temp_tweet,140);
	}
	input_tweet(text,tweet);

	int width=(signed char)tweet[0];
	bool color = width>0;
	width = abs(width);
	int height=tweet[1];

	if ( color ){
		printf("%dx%d luminance samples, %dx%d chroma samples\n",width,height,width/3,height/3);
	} else {
		printf("%dx%d luminance samples, greyscale image\n",width,height);
	}
	
	unsigned char * buffer = new unsigned char[512*512*3];
	unsigned char * buffer2 = new unsigned char[512*512*3];
	unsigned char * buffer3 = new unsigned char[512*512*3];


	init_range_decode(tweet+2,16);
	
	for ( int a=0;a<width*height;a++){
		buffer[a*3]=adaptive_range_decode();
	}
	restore(buffer,width,height);

	if ( color ){
		for ( int a=0;a<(width/3)*(height/3);a++){
			buffer2[a*3+1]=adaptive_range_decode();
			buffer2[a*3+2]=adaptive_range_decode();
		}
		restore(buffer2,width/3,height/3);
		enlarge(buffer2,buffer3,width/3,height/3,width,height);
		for ( int a=0;a<width*height;a++){
			buffer[a*3+1]=buffer3[a*3+1];
			buffer[a*3+2]=buffer3[a*3+2];
		}
	} else {
		for ( int a=0;a<width*height;a++){
			buffer[a*3+1]=128;
			buffer[a*3+2]=128;
		}
	}
	YUV2RGB(buffer,buffer2,width,height);

	while ( width*2<=128 && height*2<=128){
		double_res(buffer2,width,height);
		width*=2;
		height*=2;
	}
	if ( width >= height ){
		double d=width;
		d/=height;
		enlarge(buffer2,buffer,width,height,d*128,128);
		WriteBmp(buffer,image_file,d*128,128);
	} else {
		double d=height;
		d/=width;
		enlarge(buffer2,buffer,width,height,128,d*128);
		WriteBmp(buffer,image_file,128,d*128);
	}
	
	

}

void main(int c,char**cmd){

	char * usage = "Usage:\n-c [filename]\tCompress a 24bit/pixel .bmp image to a twitter message\n-d [tweet filename] [image filename]\tDecompress a twitter-encode image to a .bmp image";

	if ( c >= 3){
		if ( strcmp(cmd[1],"-c") ==0 || strcmp(cmd[1],"-C")==0 ){
			compress(cmd[2]);
			//decompress(NULL,"C:\\Documents and Settings\\greenwbr\\Desktop\\test.bmp");
			return;
		} else if ( c>=4 && (strcmp(cmd[1],"-d") ==0 || strcmp(cmd[1],"-D")==0 )){
			decompress(cmd[2],cmd[3]);
			return;
		}
	}
	printf(usage);
	return;
}